#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <MathUtility.h>
#include <GetTime.h>
#include <ADT.h>

#define PRINT_ADT 0

/*-----------------------------------------------------------------------*/
void Compute_Root_BoundingBox(int dim,int Npts,ADT_PT **Pts,double *a1,double *b1)
{

  int i, n;

  for (i=0;i<dim;i++) {
    a1[i] = 1.e99;
    b1[i] = -1.e99;
  }

  for (i=0;i<dim;i++) {
    for (n=0;n<Npts;n++) {      
      a1[i] = MIN(a1[i],Pts[n]->x[i]);
      b1[i] = MAX(b1[i],Pts[n]->x[i]);
    }
  }

  for (i=0;i<dim;i++) {a1[i] -= 0.1; b1[i] += 0.1;}

}
/*-----------------------------------------------------------------------*/
ADT_TREE * Create_ADT_TREE(int dim,int Npts,ADT_PT **Pts,double *a1,double *b1)
{

  int         numS, depth;

  ADT_TREE ** listS, ** listS0, * R;

  R = Create_ADT_root(dim,a1,b1,Npts,Pts);

  listS  = (ADT_TREE **) calloc(Npts,sizeof(ADT_TREE *));
  listS0 = (ADT_TREE **) calloc(Npts,sizeof(ADT_TREE *));

  numS     = 1;
  listS[0] = R;
  depth    = 0;

  do {
    depth ++;
    Create_new_subtrees(&numS,listS,listS0,dim);
  } while (numS != 0);

  qsort(Pts, Npts, sizeof(ADT_PT *), CompareNo_ADT_PT);
  
  free(listS);
  free(listS0);

  return R;

}
/*-----------------------------------------------------------------------*/
ADT_TREE * Create_ADT_root(int dim,double *a1,double *b1,int Npts,ADT_PT **Pts)
{

  int        i;

  ADT_TREE * R;

  R = Alloc_ADT_TREE(1,NULL);

  for (i=0;i<dim;i++) {
    R->a[i] = a1[i];
    R->b[i] = b1[i];
  }

  R->Pt   = Pts[0];
  R->Npts = Npts-1;
  R->Pts  = &Pts[1];

  if (PRINT_ADT == 1) Print_Subtree(R,dim,"root");

  return R;

}
/*-----------------------------------------------------------------------*/
void Create_new_subtrees(int *numS,ADT_TREE **listS,ADT_TREE **listS0,int dim)
{

  int n, numS0;

  numS0  = *numS;

  for (n=0;n<numS0;n++) listS0[n] = listS[n];

  *numS = 0;

  if (PRINT_ADT == 1) printf(" Depth = %d\n",listS0[0]->depth);

  /* printf(" -> %d Active elements\n",numS0); */

  for (n=0;n<numS0;n++) {
    Divide_subtree(listS0[n],dim);
    Check_subtree(listS0[n],numS,listS);
  }

  /* printf(" <- %d Active elements\n",*numS); */

}
/*-----------------------------------------------------------------------*/
void Divide_subtree(ADT_TREE *S,int dim)
{

  int        i, j, jmid;
  double     xj;

  if (S->Npts == 1) {

    S->son[0]       = Alloc_ADT_TREE(1,S);
    S->son[0]->Pt   = S->Pts[0];
    S->son[0]->Npts = 0;

    for (i=0;i<dim;i++) {
      S->son[0]->a[i] = S->a[i];
      S->son[0]->b[i] = S->b[i];
    }

    if (PRINT_ADT == 1) Print_Subtree(S->son[0],dim,"left son alone");

  } else {

    j = 1 + S->depth % dim;

    j --;

    switch (j) {
    case 0: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare0_ADT_PT); break;
    case 1: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare1_ADT_PT); break;
    case 2: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare2_ADT_PT); break;
    case 3: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare3_ADT_PT); break;
    case 4: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare4_ADT_PT); break;
    case 5: qsort(S->Pts,S->Npts,sizeof(ADT_PT *),Compare5_ADT_PT); break;
    default: fprintf(stderr," *** ERROR: case j=%d not implemented\n",j);
    }

    jmid = S->Npts / 2;

    xj = 0.5 * (S->Pts[jmid-1]->x[j] + S->Pts[jmid]->x[j]);

    S->son[0]       = Alloc_ADT_TREE(1,S);
    S->son[0]->Pt   = S->Pts[0];
    S->son[0]->Npts = jmid-1;
    if (S->son[0]->Npts >= 1) S->son[0]->Pts  = &(S->Pts[1]);

    S->son[1]       = Alloc_ADT_TREE(1,S);
    S->son[1]->Pt   = S->Pts[jmid];
    S->son[1]->Npts = S->Npts - S->son[0]->Npts - 2;
    if (S->son[1]->Npts >= 1) S->son[1]->Pts  = &(S->Pts[jmid+1]);

    for (i=0;i<dim;i++) {
      if (i == j) {
	S->son[0]->a[i] = S->a[i];
	S->son[0]->b[i] = xj;
	S->son[1]->a[i] = xj;
	S->son[1]->b[i] = S->b[i];
      } else {
	S->son[0]->a[i] = S->a[i];
	S->son[0]->b[i] = S->b[i];
	S->son[1]->a[i] = S->a[i];
	S->son[1]->b[i] = S->b[i];
      }
    }

    if (PRINT_ADT == 1) {
      Print_Subtree(S->son[0],dim,"left son");
      Print_Subtree(S->son[1],dim,"right son");
      printf(" Axis %d %e\n",j,xj);
    }

  }
  
}
/*-----------------------------------------------------------------------*/
void Check_subtree(ADT_TREE *S,int *numS,ADT_TREE **listS)
{

  int k;

  for (k=0;k<2;k++) {
    if (S->son[k] != NULL) {
      if (S->son[k]->Npts > 0) listS[(*numS) ++] = S->son[k];
    }
  }

}
/*-----------------------------------------------------------------------*/
void Pts_in_target_R(ADT_TREE *S,double *Tmin,double *Tmax,int dim,
		     int *numP,ADT_PT **listP)
{
  
  int i, k, test;

  test = 0;

  for (i=0;i<dim;i++) {
    if (S->Pt->x[i] > Tmax[i]) {test = 1; break;}
    if (S->Pt->x[i] < Tmin[i]) {test = 1; break;}
  }

  if (test == 0) listP[(*numP) ++] = S->Pt;

  for (k=0;k<2;k++) {
    if (S->son[k] != NULL) {
      if (S_overlap_target(S->son[k],Tmin,Tmax,dim) == 1) 
	Pts_in_target_R(S->son[k],Tmin,Tmax,dim,numP,listP);
    }
  }

}
/*-----------------------------------------------------------------------*/
void Pts_in_target_NR(ADT_TREE *R,double *Tmin,double *Tmax,int dim,
		      int *numP,ADT_PT **listP,ADT_TREE **stack)
{
  
  int        i, stackSize, finalExit, test;

  ADT_TREE * S;

  S = R;

  stackSize = 0;
  finalExit = 0;

  do {

    do {

      test = 0;

      for (i=0;i<dim;i++) {
	if (S->Pt->x[i] > Tmax[i]) {test = 1; break;}
	if (S->Pt->x[i] < Tmin[i]) {test = 1; break;}
      }

      if (test == 0) listP[(*numP) ++] = S->Pt;

      if (S->son[1] != NULL) {
	if (S_overlap_target(S->son[1],Tmin,Tmax,dim) == 1) 
	  stack[stackSize ++] = S->son[1];
      }   
      
      if (S->son[0] != NULL) {
	if (S_overlap_target(S->son[0],Tmin,Tmax,dim) == 1) S = S->son[0];
	else S = NULL;
      } else {
	S = NULL;
      }

    } while (S != NULL);

    if (stackSize == 0) finalExit = 1;
    else S = stack[-- stackSize];

  } while (finalExit == 0);


}
/*-----------------------------------------------------------------------*/
int S_overlap_target(ADT_TREE *S,double *Tmin,double *Tmax,int dim)
{

  int i;

  for (i=0;i<dim;i++) {
    if (S->a[i] > Tmax[i]) return 0;
    if (S->b[i] < Tmin[i]) return 0;
  }

  return 1;

}
/*-----------------------------------------------------------------------*/
ADT_PT * Alloc_ADT_PT(int no,double *x,int dim)
{

  int      i;

  ADT_PT * P;

  P = (ADT_PT *) calloc(1,sizeof(ADT_PT));

  if (!P) {printf(" Error in Alloc_ADT_PT\n"); exit(1);}

  P->no = no;
  P->ptr = -1;

  for (i=0;i<dim;i++) P->x[i] = x[i];

  return P;

}
/*-----------------------------------------------------------------------*/
ADT_TREE * Alloc_ADT_TREE(int no,ADT_TREE * father)
{

  ADT_TREE * S;

  S = (ADT_TREE *) calloc(1,sizeof(ADT_TREE));

  if (!S) {printf(" Error in Alloc_ADT_TREE\n"); exit(1);}

  S->no      = no;
  S->father  = father;

  if (father == NULL) S->depth = 0;
  else S->depth = father->depth + 1;

  S->son[0] = NULL;
  S->son[1] = NULL;

  return S;

}
/*-----------------------------------------------------------------------*/
int Compare0_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[0] > (*Pt2)->x[0]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int Compare1_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[1] > (*Pt2)->x[1]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int Compare2_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[2] > (*Pt2)->x[2]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int Compare3_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[3] > (*Pt2)->x[3]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int Compare4_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[4] > (*Pt2)->x[4]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int Compare5_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->x[5] > (*Pt2)->x[5]) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
int CompareNo_ADT_PT(const void *pt1,const void *pt2)
{

  ADT_PT ** Pt1, ** Pt2;

  Pt1 = (ADT_PT **) pt1;
  Pt2 = (ADT_PT **) pt2;

  if ((*Pt1)->no > (*Pt2)->no) return 1;
  else return -1;

}
/*-----------------------------------------------------------------------*/
void Print_Subtree(ADT_TREE *S,int dim,char *string)
{

  int i;

  printf(" a1 = [");
  for (i=0;i<dim;i++) printf(" %f",S->a[i]);  
  printf("] \n b1 = [");
  for (i=0;i<dim;i++) printf(" %f",S->b[i]);  
  printf("] %d (%s)\n",S->Pt->no,string);

}
/*-----------------------------------------------------------------------*/
void SetUp_ADT_PTS()
{

  int         i, Npts, dim = 2, numP;

  double      x[100][2], Tmin[4], Tmax[4], a1[4], b1[4];

  ADT_PT   ** Pts, ** listP;
  ADT_TREE  * R;

  Npts = 6;

  x[0][0] = 1.; x[0][1] = 1.;
  x[1][0] = 2.; x[1][1] = 1.;
  x[2][0] = 1.; x[2][1] = 2.;
  x[3][0] = 1.5; x[3][1] = 2.;
  x[4][0] = 2.5; x[4][1] = 2.;
  x[5][0] = 3.; x[5][1] = 2.;

  Tmin[0] = 0.; Tmax[0] = 1.9;
  Tmin[1] = 0.; Tmax[1] = 2.;

  Pts = (ADT_PT **) calloc(Npts,sizeof(ADT_PT *));

  for (i=0;i<Npts;i++) Pts[i] = Alloc_ADT_PT(i,&x[i][0],dim);

  Compute_Root_BoundingBox(dim,Npts,Pts,a1,b1);

  R = Create_ADT_TREE(dim,Npts,Pts,a1,b1);

  listP = (ADT_PT **) calloc(Npts,sizeof(ADT_PT *));

  numP = 0;

  Pts_in_target_R(R,Tmin,Tmax,dim,&numP,listP);

  printf(" -> %d :",numP);
  for (i=0;i<numP;i++) printf(" %d",listP[i]->no);
  printf("\n");

}
/*-----------------------------------------------------------------------*/
