#include <CutPlane.h>
#include <MathUtility.h>

//-------------------------------------------------------------------------------

CutPlane::CutPlane(Vec3D &_a, Vec3D &_b, Vec3D &_c) : a(_a), b(_b), c(_c) 
{

  normal = 0.5 * ((c - a) ^ (b - a));
  normal *= 1.0 / sqrt(normal*normal);

}

//-------------------------------------------------------------------------------

void CutPlane::computeBoundingBox(Vec3D meshbb[2])
{

  int i, k;

  for (i=0; i<3; ++i) {
    bb[0][i] = 1.e99;
    bb[1][i] = -1.e99;
  }

  for (i=0; i<3; ++i) {
    bb[0][i] = MIN(bb[0][i], a[i]);
    bb[0][i] = MIN(bb[0][i], b[i]);
    bb[0][i] = MIN(bb[0][i], c[i]);

    bb[1][i] = MAX(bb[1][i], a[i]);
    bb[1][i] = MAX(bb[1][i], b[i]);
    bb[1][i] = MAX(bb[1][i], c[i]);
  }


  Vec3D x[8];

  x[0][0] = meshbb[0][0];
  x[0][1] = meshbb[0][1];
  x[0][2] = meshbb[0][2];

  x[1][0] = meshbb[0][0];
  x[1][1] = meshbb[0][1];
  x[1][2] = meshbb[1][2];

  x[2][0] = meshbb[1][0];
  x[2][1] = meshbb[0][1];
  x[2][2] = meshbb[1][2];

  x[3][0] = meshbb[1][0];
  x[3][1] = meshbb[0][1];
  x[3][2] = meshbb[0][2];

  x[4][0] = meshbb[0][0];
  x[4][1] = meshbb[1][1];
  x[4][2] = meshbb[0][2];

  x[5][0] = meshbb[0][0];
  x[5][1] = meshbb[1][1];
  x[5][2] = meshbb[1][2];

  x[6][0] = meshbb[1][0];
  x[6][1] = meshbb[1][1];
  x[6][2] = meshbb[1][2];

  x[7][0] = meshbb[1][0];
  x[7][1] = meshbb[1][1];
  x[7][2] = meshbb[0][2];


  static int edgeEnd[12][2] = { {0,3}, {1,2}, {5,6}, {4,7},   // along x-axis
				{0,4}, {1,5}, {2,6}, {3,7},   // along y-axis
				{0,1}, {3,2}, {7,6}, {4,5} }; // along z-axis

  Vec3D p[12];
  int cut = 0;

  for (i=0; i<3; ++i) {
    for (k=0; k<4; ++k) {

      Vec3D d = x[ edgeEnd[4*i+k][0] ];
      Vec3D e = x[ edgeEnd[4*i+k][1] ];

      int cross = checkPiercePoint(d, e);

      if (cross == 1 || cross == 2 || cross == 3) {
	double s = constructPiercePoint(d, e);
	p[cut] = d + s * (e - d);
	if (p[cut][i] >= meshbb[0][i] && p[cut][i] <= meshbb[1][i])
	  ++cut;
      }

    }
  }

  for (i=0; i<3; ++i) {
    for (k=0; k<cut; ++k) {
      bb[0][i] = MIN(bb[0][i], p[k][i]);
      bb[1][i] = MAX(bb[1][i], p[k][i]);
    }
  }

  fprintf(stdout, " ... mesh bb: (%e %e %e)\n", meshbb[0][0], meshbb[0][1], meshbb[0][2]);
  fprintf(stdout, "              (%e %e %e)\n", meshbb[1][0], meshbb[1][1], meshbb[1][2]);

  fprintf(stdout, " ... slice bb: (%e %e %e)\n", bb[0][0], bb[0][1], bb[0][2]);
  fprintf(stdout, "               (%e %e %e)\n", bb[1][0], bb[1][1], bb[1][2]);

}

//-------------------------------------------------------------------------------
