#ifndef _DOMAIN_H_
#define _DOMAIN_H_

#include <Node.h>
#include <Edge.h>
#include <Face.h>
#include <Tet.h>
#include <ADT.h>
#include <PostOperator.h>

#include <string>
#include <stack>
using namespace std;

class CutPlane;
class PostOp;

template<class Scalar> class Vec;
template<class Scalar, int dim> class SVec;

//-------------------------------------------------------------------------------

class Domain {

  char *name;

  NodeSet nodes;
  EdgeSet edges;
  FaceSet faces;
  TetSet tets;

  Vec3D bb[2];

  PostOp *postop;

  int solutionDimension; // TG scalar == 1 - vector == 3
public:

  Domain(char *_name) : name(_name), nodes(0) { postop = 0; solutionDimension = 1; }
  ~Domain() {}

  void readGeometry(char *, char *, int);
  void readAsciiGeometry(char *);
  void readBinaryGeometry(char *, char *);
  void readXpostGeometry(char *);
  bool readSolution(char *, FILE *&, PostOp::TypeSol type, int, int, double &, double*);
  bool readPosition(char *, int, double&, SVec<double,3> &);
  void writeSolution(char *, SVec<double,5> &);
  void writeGmshGeometry(Vec<double> &);

  void numberEdges();

  void computeBoundingBox(SVec<double,3> &, bool = false);
  void linecuts(char*, char*, char*, CutLine*);
  void makelinecut(char*, CutLine&, SVec<double,3> &X, Vec<CutLinePt>&);
  void planecuts(char*, char*, char*, CutPlane*, int, int, int);
  void makeCut(char*, CutPlane &, SVec<double,3> &, SVec<double,3> &, 
	       SVec<int,4> &, SVec<int,3> &, Vec<double>&, Vec<int>&, Vec<int>&);
  void writeCut(FILE*, FILE*, double,  SVec<double,3> &, Vec<double> &, SVec<double,3> &,
		SVec<int,4> &, SVec<int,3> &, Vec<double> &, Vec<int>&, Vec<int>&);
  void constructTraceOnSurface(char *, char*, SVec<double,3> &, Vec<double> &,
			       SVec<int,4> &, SVec<int,3> &, Vec<double> &);

  ADT_TREE *setupADT(int &, ADT_PT **&);
  void locatePoints(ADT_TREE *, int);
  int interpolateSolution(int, ADT_PT **, SVec<double,5> &, SVec<double,5> &);
  void checkOutsidePoints(int, ADT_PT **, SVec<double,5> &);
  void initializeOutsidePoints(int, ADT_PT**, NodeSet&, SVec<double,5> &, SVec<double,5> &);
  int createNodeToNodeConnectivity(int *&, int *&);

  int getNumNodes() { return nodes.size(); }
  NodeSet& getNodes() { return nodes; }

  void nodalValueExtract(char*, char*, char*, int);
  void pointWiseValueExtract(char* nameSol, char* namePos, char* suffix, double sensor[3]);

};

//-------------------------------------------------------------------------------

#endif

