#include <CutPlane.h>
#include <Domain.h>
#include <ADT.h>

#include <stdlib.h>
#include <stdio.h>
#include <stack>
#include <string>
using std::stack;

#define MAXLINE 500

bool gmsh;
bool symmetry;
double xyzsym[3];

//-------------------------------------------------------------------------------

int readCommandLineCuts(int argc, char **argv, char *conname, char *geoname, 
			char *solname, char *posname, char *suffix, 
			CutLine*& cutLine, CutPlane *&cutPlane, int &freq, int &minPlane,
                        int &maxPlane, int &nodenum, double *&sensor)
{

  int j = 0;
  int typeRead = 2;
  int typeSol = 0;
  int typePos = 0;
  int typeSuffix = 0;

  for (int i=1; i<argc; ++i) {

    if (strcmp(argv[i],"-con") == 0) { 
      strcpy(conname, argv[i+1]); 
      typeRead = 1; 
    }
    else if (strcmp(argv[i],"-mesh") == 0) { 
      strcpy(geoname, argv[i+1]); 
      ++j; 
    }
    else if (strcmp(argv[i],"-res") == 0) { 
      strcpy(solname, argv[i+1]); 
      typeSol = 1;
    }
    else if (strcmp(argv[i],"-disp") == 0) { 
      strcpy(posname, argv[i+1]); 
      typePos = 1;
    }
    else if (strcmp(argv[i],"-sinus") == 0) { 
      typeRead = 0; 
    }
    else if (strcmp(argv[i],"-suffix") == 0) { 
      typeSuffix = 1;
      strcpy(suffix, argv[i+1]);
    }
    else if (strcmp(argv[i],"-plane") == 0) { 
      Vec3D x[3];
      for (int k=0; k<3; ++k) 
        for (int l=0; l<3; ++l)
          x[k][l] = atof(argv[i+1 + 3*k + l]);

      cutPlane = new CutPlane(x[0], x[1], x[2]);
      ++j;
    }
    else if (strcmp(argv[i],"-line") == 0) { 
      Vec3D x[2];
      for (int k=0; k<2; ++k) 
        for (int l=0; l<3; ++l)
          x[k][l] = atof(argv[i+1 + 3*k + l]);

      cutLine = new CutLine(x[0], x[1]);
      ++j;
    }
    else if (strcmp(argv[i],"-freq") == 0) { 
      freq = atoi(argv[i+1]);
      fprintf(stdout, "frequency of postprocessing is %d\n", freq);
    }
    else if (strcmp(argv[i],"-min") == 0) { 
      minPlane = atoi(argv[i+1]);
    }
    else if (strcmp(argv[i],"-max") == 0) { 
      maxPlane = atoi(argv[i+1]);
    }
    else if (strcmp(argv[i],"-nodenum") == 0) { 
      nodenum = atoi(argv[i+1]);
      fprintf(stdout, "node number to be postprocessed is %d\n", nodenum);
      ++j;
    }
    else if(strcmp(argv[i], "-sensor") == 0) {
      sensor = new double[3];
      for (int k=0; k<3; k++)
        sensor[k] = atof(argv[i+1+k]);
      fprintf(stdout, "sensor location is (%e, %e, %e)\n", sensor[0], sensor[1], sensor[2]);
      ++j;
    }

  }

  //if (j != 2) {
  //  fprintf(stdout, "usage: meshtools -mesh <filename> [-con <filename>]\n");
  //  fprintf(stdout, "                 [-res <filename>] [-disp <filename>]\n");
  //  fprintf(stdout, "                 [-suffix <name>] [-line <x0 y0 z0 x1 y1 z1>]\n");
  //  fprintf(stdout, "                 -plane <x0 y0 z0 x1 y1 z1 x2 y2 z2>\n");
  //  exit(1);
  //}

  if (j != 2) {
	  fprintf(stdout, "Meshtools Usage               (c) Stanford University        \n");
		fprintf(stdout, "meshtools       -mesh <filename>    Mesh File           \n");
		fprintf(stdout, "\t\t[-con <filename>]   Connectivity File         \n");
    fprintf(stdout, "\t\t[-res <filename>]   Result File              \n");
		fprintf(stdout, "\t\t[-disp <filename>]  Displacement File         \n");
 		fprintf(stdout, "\t\t[-suffix <name>]    Suffix to append to output files \n");
 		fprintf(stdout, "\t\t[-freq <value>]     Output frequency for plane cut\n");
 		fprintf(stdout, "\t\t[-min <value>]      Minimum output time slice for plane cut\n");
 		fprintf(stdout, "\t\t[-max <value>]      Maximum output time slice for plane cut\n");
		fprintf(stdout, "\t\t[-line <x0 y0 z0 x1 y1 z1>] Reference Cut Line \n");
	  fprintf(stdout, "\t\t-plane <x0 y0 z0 x1 y1 z1 x2 y2 z2> Reference Cut Plane\n");
 		fprintf(stdout, "\t\t[-nodenum <node number>]  Nodal value\n");
 		fprintf(stdout, "\t\t[-sensor <x y z>]   Sensor location\n");
		exit(1);
  }
  
  
  if (typeRead != 1)
   // conname = "";
    strcpy(conname, ""); //HB

  if (typeSol == 0)
   // solname = "";
    strcpy(solname, ""); //HB

  if (typeSuffix == 0)
   // suffix = "";
    strcpy(suffix, ""); //HB

  if (typePos == 0)
   // posname = "";
    strcpy(posname, ""); //HB

  return typeRead;

}

//-------------------------------------------------------------------------------

void performCuts(int argc, char** argv)
{

  char ncon[MAXLINE], ngeo[MAXLINE], nsol[MAXLINE], npos[MAXLINE], suffix[MAXLINE];
  int freq = 1;
  int globNodeNum = 0;
  int minplane = -1;
  int maxplane = -1;

  double *sensor = 0;
  CutLine *line = 0;
  CutPlane *plane = 0;

  int typeRead = readCommandLineCuts(argc, argv, ncon, ngeo, nsol, npos, suffix, 
                   line, plane, freq, minplane, maxplane, globNodeNum, sensor);

  Domain domain(ngeo);

  domain.readGeometry(ncon, ngeo, typeRead);
  fprintf(stdout, "Mesh and Connectivity have been read\n");

  if (line)
    domain.linecuts(nsol, npos, suffix, line);

  if (plane)
    domain.planecuts(nsol, npos, suffix, plane, freq, minplane, maxplane);

  if (globNodeNum>0){
    fprintf(stdout, "...Extracting nodal values...\n");
    domain.nodalValueExtract(nsol, npos, suffix, globNodeNum);
  }

  if (sensor){
    fprintf(stdout, "...Extracting sensor values...\n");
    domain.pointWiseValueExtract(nsol, npos, suffix, sensor);
  }

}

//-------------------------------------------------------------------------------

int main(int argc, char** argv)
{

  performCuts(argc, argv);

  return 0;
 
}

//-------------------------------------------------------------------------------
