#ifndef _NODE_H_
#define _NODE_H_

#include <stdio.h>

#include <Vector.h>
#include <Vector3D.h>
#include <BinFileHandler.h>

//-------------------------------------------------------------------------------

struct CutLinePt {

  int nodes[3];
  double r, t;

  CutLinePt() {}
  ~CutLinePt() {}

  int& operator[](int i) { return nodes[i]; }

};

//-------------------------------------------------------------------------------

struct CutPt {

  int i, j;
  double s;

  CutPt() {}
  ~CutPt() {}

};

//------------------------------------------------------------------------------

class NodeSet : public SVec<double,3> {

public:

  NodeSet(int n) : SVec<double,3>(n) {}
  ~NodeSet() {}

  void read(int, FILE *);
  void read(int, double (*)[3]);
  void readXpost(int, FILE *);

};

//-------------------------------------------------------------------------------

inline
void NodeSet::read(int _numNodes, FILE *fptr)  
{

  resize(_numNodes);

  char line[MAXLINE];
  for (int i=0; i<_numNodes; ++i) {
    fgets(line, MAXLINE, fptr);
    sscanf(line, "%lf %lf %lf", &v[i][0], &v[i][1], &v[i][2]);
  }

  fprintf(stdout, "       %d nodes\n", _numNodes);

}

//-------------------------------------------------------------------------------

inline
void NodeSet::readXpost(int _numNodes, FILE *fptr)  
{

  resize(_numNodes);

  char line[MAXLINE];
  for (int i=0; i<_numNodes; ++i) {
    fgets(line, MAXLINE, fptr);
    sscanf(line, "%*s %lf %lf %lf", &v[i][0], &v[i][1], &v[i][2]);
  }

  fprintf(stdout, "       %d nodes\n", _numNodes);

}

//-------------------------------------------------------------------------------

inline
void NodeSet::read(int _numNodes, double (*X)[3])  
{

  resize(_numNodes);

  for (int i=0; i<_numNodes; ++i)
    for (int k=0; k<3; ++k)
      v[i][k] = X[i][k];

}

//-------------------------------------------------------------------------------

#endif
