#ifndef _TET_H_
#define _TET_H_

#include <stdio.h>

#include <ADT.h>
#include <Node.h>
#include <CutPlane.h>
#include <MathUtility.h>

class Face;
class EdgeSet;
class BinFileHandler;
struct CutPt;
template <class Scalar> class Vec;
template <class Scalar, int dim> class SVec;

//-------------------------------------------------------------------------------

class Tet {

public:

  static const int face_def[4][3];
  static const int face2node_def[4];

private:

  int nodeNum[4];
  //PHG int edgeNum[6];

  Vec3D bb[2];

public:

  Tet() {}
  ~Tet() {}
  Tet(const Tet &);

  int &operator[](int i) { return nodeNum[i]; }

  void numberEdges(EdgeSet &);
  double computeVolume(SVec<double,3>&);
  void computeBoundingBox(SVec<double,3>&);

  int checkOverlap(Vec3D *);
  int makelinecut(SVec<double,3>&, CutLine&, map<int,int>&, MapEdge&, 
		  MapFace&, stack<CutLinePt>&, MapEdge&);
  int makeplanecut(SVec<double,3> &, CutPlane &, Vec<int> &, 
		   MapEdge &, MapFace &, stack<CutPt> &);

  void locatePoints(NodeSet &, int, ADT_TREE *, ADT_PT **, ADT_TREE **);
  int pointIn(SVec<double,3>&, double *);
  int pointInRobust(SVec<double,3>&, double *);
  double computeGradientP1Function(SVec<double,3>&, double [4][3]);

  template<int dim>
  void interpolateSolution(SVec<double,3>&, SVec<double,dim> &, double *, double *);

  template<int dim>
  void deriveSolution(SVec<double,3>&, SVec<double,dim>&, double*, double*, double*);

  void barycentricCoord(SVec<double,3> &X, double point[3], 
                        double localCoord[3]);
  void interpolate(Vec<double> &values, int dim, double localCoord[3], 
                      double *interpolated);

};

//-------------------------------------------------------------------------------

class TetSet {

  int numTets;
  Tet *tets;
  
public:

  TetSet() { numTets = 0; tets = 0; }
  ~TetSet() { if (tets) delete[] tets; }

  Tet &operator[](int i) const { return tets[i]; }

  void read(int, FILE *);
  void read(int, int (*)[4]);
  void readXpost(int, FILE *);
  void numberEdges(EdgeSet &);

  int size() const { return numTets; }
  void resize(int n) { numTets = n; tets = new Tet[numTets]; }
  
};

//-------------------------------------------------------------------------------

inline
void Tet::computeBoundingBox(SVec<double,3>& X)
{

  int i, k;

  for (k=0; k<3; ++k) {
    bb[0][k] = X[ nodeNum[0] ][k];
    bb[1][k] = bb[0][k];
  }

  for (i=1; i<4; ++i) {
    for (k=0; k<3; ++k) {
      bb[0][k] = MIN(bb[0][k], X[ nodeNum[i] ][k]);
      bb[1][k] = MAX(bb[1][k], X[ nodeNum[i] ][k]);
    }
  }

}

//-------------------------------------------------------------------------------

inline
int Tet::checkOverlap(Vec3D *elembb)
{

  for (int i=0; i<3; ++i) {
    if (bb[0][i] > elembb[1][i]) return 0;
    if (bb[1][i] < elembb[0][i]) return 0;
  }

  return 1;

}

//-------------------------------------------------------------------------------

#ifdef TEMPLATE_FIX
#include <Tet.C>
#endif

#endif
