#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <Edge.h>
#include <Face.h>
#include <Tet.h>
#include <Vector.h>
#include <BinFileHandler.h>

#define MAXLINE 500

extern double orient3d(double *, double *, double *, double *);

const int Tet::face_def[4][3] = { {0,1,2}, {0,3,1}, {0,2,3}, {1,3,2} };
const int Tet::face2node_def[4] = {3, 2, 1, 0};

//-------------------------------------------------------------------------------

Tet::Tet(const Tet &tet)
{
  
  nodeNum[0] = tet.nodeNum[0];
  nodeNum[1] = tet.nodeNum[1];
  nodeNum[2] = tet.nodeNum[2];
  nodeNum[3] = tet.nodeNum[3];

  /* PHG
  edgeNum[0] = tet.edgeNum[0];
  edgeNum[1] = tet.edgeNum[1];
  edgeNum[2] = tet.edgeNum[2];
  edgeNum[3] = tet.edgeNum[3];
  edgeNum[4] = tet.edgeNum[4];
  edgeNum[5] = tet.edgeNum[5];
  */

}

//-------------------------------------------------------------------------------

void Tet::numberEdges(EdgeSet &edges)
{

  /* PHG
  edgeNum[0] = edges.find(nodeNum[0], nodeNum[1]);
  edgeNum[1] = edges.find(nodeNum[0], nodeNum[2]);
  edgeNum[2] = edges.find(nodeNum[0], nodeNum[3]);
  edgeNum[3] = edges.find(nodeNum[1], nodeNum[2]);
  edgeNum[4] = edges.find(nodeNum[1], nodeNum[3]);
  edgeNum[5] = edges.find(nodeNum[2], nodeNum[3]);
  */

}

//-------------------------------------------------------------------------------

double Tet::computeVolume(SVec<double,3>& X)
{

  static double sixth = 1.0/6.0;

  Vec3D x[4] = {X[ nodeNum[0] ], X[ nodeNum[1] ], X[ nodeNum[2] ], X[ nodeNum[3] ]};

  Vec3D v1 = x[1] - x[0];
  Vec3D v2 = x[2] - x[0];
  Vec3D v3 = x[3] - x[0];

  return sixth * (v3 * (v1 ^ v2));

}

//-------------------------------------------------------------------------------

int Tet::makelinecut(SVec<double,3>& X, CutLine& line, map<int,int>& lcnodes, 
		     MapEdge& lcedges, MapFace& lcfaces, stack<CutLinePt>& pts, MapEdge& seg)
{

  static const int edge2node[6][2] = { {0,1}, {1,2}, {0,2}, {1,3}, {0,3}, {2,3} };
  static const int face2node[4][3] = { {0,1,2}, {0,3,1}, {0,2,3}, {1,3,2} };
  static const int face2edge[4][3] = { {0,1,2}, {4,3,0}, {2,5,4}, {3,5,1} };

  Vec3D x[4] = {X[ nodeNum[0] ], X[ nodeNum[1] ], X[ nodeNum[2] ], X[ nodeNum[3] ]};

  int pierceFace[4], pierceEdge[6], pierceNode[4];
  int l;
  for (l=0; l<6; ++l) {
    pierceEdge[l] = -1; 
    if (l < 4) {
      pierceFace[l] = -1;  
      pierceNode[l] = -1; 
    }
  }

  int pierce;
  for (l=0; l<4; ++l) {
    pierce = line.checkPiercePoint(x[face2node[l][0]], x[face2node[l][1]], x[face2node[l][2]]);
    if (pierce == 0) 
      pierceFace[l] = l;
    else if (pierce == 1) 
      pierceEdge[face2edge[l][0]] = l;
    else if (pierce == 2) 
      pierceEdge[face2edge[l][1]] = l;
    else if (pierce == 3) 
      pierceEdge[face2edge[l][2]] = l;
    else if (pierce == -1) 
      pierceNode[face2node[l][0]] = l;
    else if (pierce == -2) 
      pierceNode[face2node[l][1]] = l;
    else if (pierce == -3) 
      pierceNode[face2node[l][2]] = l;
  }

  pierce = 0;
  for (l=0; l<6; ++l) {
    if (pierceEdge[l] != -1)
      ++pierce;
    if (l < 4) {
      if (pierceFace[l] != -1)
	++pierce;
      if (pierceNode[l] != -1)
	++pierce;
    }
  }

  if (pierce != 2) return 0;

  pair<int,int> newseg(-1, -1);

  for (l=0; l<6; ++l) {
    if (pierceEdge[l] != -1) {
      int noEdge;
      pair<int,int> p(nodeNum[ edge2node[l][0] ], nodeNum[ edge2node[l][1] ]);
      if (p.first > p.second) 
	swap(p.first, p.second);
      MapEdge::iterator it = lcedges.find(p);
      if (it == lcedges.end()) {
	CutLinePt pt;
	pt[0] = nodeNum[ face2node[pierceEdge[l]][0] ];
	pt[1] = nodeNum[ face2node[pierceEdge[l]][1] ];
	pt[2] = nodeNum[ face2node[pierceEdge[l]][2] ];
	line.constructPiercePoint(x[face2node[pierceEdge[l]][0]], x[face2node[pierceEdge[l]][1]], 
				  x[face2node[pierceEdge[l]][2]], 0, &pt.r, &pt.t);
	noEdge = pts.size();
	lcedges[p] = noEdge;
	pts.push(pt);
      }
      else
	noEdge = it->second;

      if (newseg.first == -1)
	newseg.first = noEdge;
      else
	newseg.second = noEdge;
    }
    if (l < 4) {
      if (pierceNode[l] != -1) {
	int noNode;
	map<int,int>::iterator it = lcnodes.find(nodeNum[l]);
	if (it == lcnodes.end()) {
	  CutLinePt pt;
	  pt[0] = nodeNum[ face2node[pierceNode[l]][0] ];
	  pt[1] = nodeNum[ face2node[pierceNode[l]][1] ];
	  pt[2] = nodeNum[ face2node[pierceNode[l]][2] ];
	  line.constructPiercePoint(x[face2node[pierceNode[l]][0]], x[face2node[pierceNode[l]][1]], 
				    x[face2node[pierceNode[l]][2]], 0, &pt.r, &pt.t);
	  noNode = pts.size();
	  lcnodes[nodeNum[l]] = noNode;
	  pts.push(pt);
	}
	else
	  noNode = it->second;

	if (newseg.first == -1)
	  newseg.first = noNode;
	else
	  newseg.second = noNode;
      }
      if (pierceFace[l] != -1) {
	int noFace;
	Face f;
	f[0] = nodeNum[ face2node[l][0] ];
	f[1] = nodeNum[ face2node[l][1] ];
	f[2] = nodeNum[ face2node[l][2] ];
	f.reorder();
	MapFace::iterator it = lcfaces.find(f);
	if (it == lcfaces.end()) {
	  CutLinePt pt;
	  pt[0] = nodeNum[ face2node[pierceFace[l]][0] ];
	  pt[1] = nodeNum[ face2node[pierceFace[l]][1] ];
	  pt[2] = nodeNum[ face2node[pierceFace[l]][2] ];
	  line.constructPiercePoint(x[face2node[pierceFace[l]][0]], x[face2node[pierceFace[l]][1]], 
				    x[face2node[pierceFace[l]][2]], 0, &pt.r, &pt.t);
	  noFace = pts.size();
	  lcfaces[f] = noFace;
	  pts.push(pt);
	}
	else
	  noFace = it->second;

	if (newseg.first == -1)
	  newseg.first = noFace;
	else
	  newseg.second = noFace;
      }
    }
  }

  if (newseg.first > newseg.second) 
    swap(newseg.first, newseg.second);
  if (seg.find(newseg) == seg.end()) {
    int num = seg.size();
    seg[newseg] = num;
  }

  return 1;

}

//-------------------------------------------------------------------------------

int Tet::makeplanecut(SVec<double,3> &X, CutPlane &cutPlane, Vec<int> &nodeCut, 
		      MapEdge &listCutEdges, MapFace &listCutFaces, stack<CutPt> &listCutPts)
{

  static int edgeEnd[6][2] = { {0,1}, {0,2}, {0,3}, {1,2}, {1,3}, {2,3} };

  Vec3D x[4] = {X[ nodeNum[0] ], X[ nodeNum[1] ], X[ nodeNum[2] ], X[ nodeNum[3] ]};

  int l;
  int pierceEdge[6], noEdges[6];
  for (l=0; l<6; ++l) { pierceEdge[l] = 0; noEdges[l] = -1; }

  int pierceNode[4], noNodes[4];
  for (l=0; l<4; ++l) { pierceNode[l] = 0; noNodes[l] = -1; }

  for (l=0; l<6; ++l) {
    int pierce = cutPlane.checkPiercePoint(x[ edgeEnd[l][0] ], x[ edgeEnd[l][1] ]);

    if (pierce == 1) 
      pierceEdge[l] = 1;
    else if (pierce == 2) 
      pierceNode[ edgeEnd[l][0] ] = 1;
    else if (pierce == 3) 
      pierceNode[ edgeEnd[l][1] ] = 1;
    else if (pierce == 4) { 
      pierceNode[ edgeEnd[l][0] ] = 1; 
      pierceNode[ edgeEnd[l][1] ] = 1; 
    }
  }

  int numPierceEdge = 0;
  for (l=0; l<6; ++l)
    if (pierceEdge[l] == 1) ++numPierceEdge;

  int numPierceNode = 0;
  for (l=0; l<4; ++l)
    if (pierceNode[l] == 1) ++numPierceNode;

  int type = 0;

  if (numPierceNode == 0 && numPierceEdge == 3)
    type = 1;
  else if (numPierceNode == 0 && numPierceEdge == 4)
    type = 2;
  else if (numPierceNode == 1 && numPierceEdge == 2)
    type = 3;
  else if (numPierceNode == 2 && numPierceEdge == 1)
    type = 4;
  else if (numPierceNode == 3 && numPierceEdge == 0)
    type = 5;

  if (type == 0) return 0;

  CutPt pt;

  for (l=0; l<6; ++l) {
    if (pierceEdge[l] == 1) {

      int first = nodeNum[ edgeEnd[l][0] ];
      int second = nodeNum[ edgeEnd[l][1] ];
      if (first > second) swap(first, second);
      pair<int,int> p(first, second);

      MapEdge::iterator ite = listCutEdges.find(p);

      if (ite == listCutEdges.end()) {
	pt.i = nodeNum[ edgeEnd[l][0] ];
	pt.j = nodeNum[ edgeEnd[l][1] ];
	pt.s = cutPlane.constructPiercePoint(x[ edgeEnd[l][0] ], x[ edgeEnd[l][1] ]);

	noEdges[l] = listCutPts.size();
	listCutEdges[p] = noEdges[l];

	listCutPts.push(pt);
      }
      else
	noEdges[l] = ite->second;
    }
  }

  for (l=0; l<4; ++l) {
    if (pierceNode[l] == 1) {
      if (nodeCut[ nodeNum[l] ] == -1) {
	pt.i = nodeNum[l];
	pt.j = pt.i;
	pt.s = 0.0;

	noNodes[l] = listCutPts.size();
	nodeCut[ nodeNum[l] ] = noNodes[l];

	listCutPts.push(pt);
      }
      else
	noNodes[l] = nodeCut[ nodeNum[l] ];
    }
  }

  int i;
  Face f, fbis;

  if (type == 1 || type == 2) {

    i = 0;
    for (l=0; l<6; ++l) {
      if (noEdges[l] != -1) {
	if (i < 3) 
	  f[i] = noEdges[l];
	else if (i == 3) { 
	  fbis[0] = f[1];
	  fbis[1] = f[2];
	  fbis[2] = noEdges[l];
	  break;
	}

	++i;
      }
    }

    listCutFaces[f] = type;
    if (type == 2) listCutFaces[fbis] = type;

  }
  else if (type == 3 || type == 4 || type == 5) {

    int i = 0;
    for (l=0; l<6; ++l)
      if (noEdges[l] != -1) f[i++] = noEdges[l];
    
    for (l=0; l<4; ++l)
      if (noNodes[l] != -1) f[i++] = noNodes[l];

    int add = 1;

    //if (type == 5 && listCutFaces.find(f) != listCutFaces.end()) add = 0;

    /*
    */
    if (type == 5) {
      f.reorder();
      MapFace::iterator itf = listCutFaces.find(f);
      if (itf != listCutFaces.end()) add = 0;
    }

    if (add == 1) 
      listCutFaces[f] = type;

  }

  return 1;
  
}

//-------------------------------------------------------------------------------

void Tet::locatePoints(NodeSet &nodes, int tetNum, ADT_TREE *rootADT, 
		       ADT_PT **ptsTmp, ADT_TREE **stack)
{

  int numPts = 0;
  //Pts_in_target_R(rootADT, Tmin, Tmax, 3, &numPts, ptsTmp);
  Pts_in_target_NR(rootADT, bb[0], bb[1], 3, &numPts, ptsTmp, stack);

  for (int i=0; i<numPts; ++i) {
    if (ptsTmp[i]->ptr == -1) {
      int ans = pointInRobust(nodes, ptsTmp[i]->x);
      if (ans == 1) 
	ptsTmp[i]->ptr = tetNum;
    }
  }

}

//-------------------------------------------------------------------------------

int Tet::pointIn(SVec<double,3>& nodes, double *pa)
{

  if (pa[0] < bb[0][0] || pa[0] > bb[1][0]) return 0;
  if (pa[1] < bb[0][1] || pa[1] > bb[1][1]) return 0;
  if (pa[2] < bb[0][2] || pa[2] > bb[1][2]) return 0;

  return pointInRobust(nodes, pa);

}

//-------------------------------------------------------------------------------

int Tet::pointInRobust(SVec<double,3>& nodes, double *pa)
{

  double* p0 = nodes[ nodeNum[0] ];
  double* p1 = nodes[ nodeNum[1] ];
  double* p2 = nodes[ nodeNum[2] ];
  double* p3 = nodes[ nodeNum[3] ];

  if (orient3d(p0, p1, p2, pa) > 0.0) return 0;
  if (orient3d(p0, p2, p3, pa) > 0.0) return 0;
  if (orient3d(p0, p3, p1, pa) > 0.0) return 0;
  if (orient3d(p1, p3, p2, pa) > 0.0) return 0;

  return 1;

}

//------------------------------------------------------------------------------

double Tet::computeGradientP1Function(SVec<double,3> &nodes, double nGrad[4][3])
{

  static double sixth = 1.0/6.0;

  double jac[3][3];

  //Jacobian
  // J_ij = dx_i/dxi_j
  jac[0][0] = nodes[ nodeNum[1] ][0] - nodes[ nodeNum[0] ][0];
  jac[0][1] = nodes[ nodeNum[2] ][0] - nodes[ nodeNum[0] ][0];
  jac[0][2] = nodes[ nodeNum[3] ][0] - nodes[ nodeNum[0] ][0];
  jac[1][0] = nodes[ nodeNum[1] ][1] - nodes[ nodeNum[0] ][1];
  jac[1][1] = nodes[ nodeNum[2] ][1] - nodes[ nodeNum[0] ][1];
  jac[1][2] = nodes[ nodeNum[3] ][1] - nodes[ nodeNum[0] ][1];
  jac[2][0] = nodes[ nodeNum[1] ][2] - nodes[ nodeNum[0] ][2];
  jac[2][1] = nodes[ nodeNum[2] ][2] - nodes[ nodeNum[0] ][2];
  jac[2][2] = nodes[ nodeNum[3] ][2] - nodes[ nodeNum[0] ][2];

  // compute determinant of jac
  double dOmega = jac[0][0] * (jac[1][1] * jac[2][2] - jac[1][2] * jac[2][1]) +
                  jac[1][0] * (jac[0][2] * jac[2][1] - jac[0][1] * jac[2][2]) +
                  jac[2][0] * (jac[0][1] * jac[1][2] - jac[0][2] * jac[1][1]);

  // compute inverse matrix of jac
  // Maple code used
  double t17 = -1.0/dOmega;

  //compute shape function gradients
  nGrad[1][0] =  (-jac[1][1] * jac[2][2] + jac[1][2] * jac[2][1] ) * t17;
  nGrad[1][1] =  ( jac[0][1] * jac[2][2] - jac[0][2] * jac[2][1] ) * t17;
  nGrad[1][2] = -( jac[0][1] * jac[1][2] - jac[0][2] * jac[1][1] ) * t17;

  nGrad[2][0] = -(-jac[1][0] * jac[2][2] + jac[1][2] * jac[2][0] ) * t17;
  nGrad[2][1] = -( jac[0][0] * jac[2][2] - jac[0][2] * jac[2][0] ) * t17;
  nGrad[2][2] =  ( jac[0][0] * jac[1][2] - jac[0][2] * jac[1][0] ) * t17;

  nGrad[3][0] = -( jac[1][0] * jac[2][1] - jac[1][1] * jac[2][0] ) * t17;
  nGrad[3][1] =  ( jac[0][0] * jac[2][1] - jac[0][1] * jac[2][0] ) * t17;
  nGrad[3][2] = -( jac[0][0] * jac[1][1] - jac[0][1] * jac[1][0] ) * t17;

  // Shape function gradients dN_i/dx_i = dN/dxi * transpose(jInv)
  // Note: 1st index = shape function #
  // 2nd index = direction (0=x, 1=y, 2=z)

  nGrad[0][0] = -( nGrad[1][0] + nGrad[2][0] + nGrad[3][0] );
  nGrad[0][1] = -( nGrad[1][1] + nGrad[2][1] + nGrad[3][1] );
  nGrad[0][2] = -( nGrad[1][2] + nGrad[2][2] + nGrad[3][2] );

  return sixth * dOmega;

}

//-------------------------------------------------------------------------------
// assumes a non-flat tet

void Tet::barycentricCoord(SVec<double,3> &X, double point[3],
                           double localCoord[3])
{

  // create base from tet's nodes
  Vec3D x[4] = {X[ nodeNum[0] ], X[ nodeNum[1] ], X[ nodeNum[2] ], X[ nodeNum[3] ]};

  Vec3D e[3] = {x[1]-x[0], x[2]-x[0], x[3]-x[0]};

  // vector to decompose on that base
  Vec3D v(point[0]-x[0][0], point[1]-x[0][1], point[2]-x[0][2]);

  // compute normals to all 3 planes opposite of first node
  Vec3D n[3] = {e[1] ^ e[2], e[2] ^ e[0], e[0] ^ e[1]};
  for (int i=0; i<3; i++){
    double invNorm = 1.0/sqrt(n[i][0]*n[i][0]+n[i][1]*n[i][1]+n[i][2]*n[i][2]);
    n[i] *= invNorm;
  }

  localCoord[0] = (v*n[0])/(e[0]*n[0]);
  localCoord[1] = (v*n[1])/(e[1]*n[1]);
  localCoord[2] = (v*n[2])/(e[2]*n[2]);
#ifdef ARL_DEBUG
    fprintf(stdout, " ... Barycentric coordinates are (%e %e %e)\n", localCoord[0],localCoord[1],localCoord[2]);
#endif

}

//-------------------------------------------------------------------------------

void Tet::interpolate(Vec<double> &values, int dim, double localCoord[3], 
                      double *interpolated)
{

  for(int j=0; j<dim; j++)
    interpolated[j] = values[nodeNum[0]*dim+j] 
          + localCoord[0]*(values[nodeNum[1]*dim+j] - values[nodeNum[0]*dim+j])
          + localCoord[1]*(values[nodeNum[2]*dim+j] - values[nodeNum[0]*dim+j])
          + localCoord[2]*(values[nodeNum[3]*dim+j] - values[nodeNum[0]*dim+j]);

#ifdef ARL_DEBUG
    fprintf(stdout, " ... Data for Interpolation = %e %e %e %e\n", values[nodeNum[0]*dim],values[nodeNum[1]*dim],values[nodeNum[2]*dim],values[nodeNum[3]*dim]);
    fprintf(stdout, " ... Interpolated value = %e\n", interpolated[0]);
#endif

}

//-------------------------------------------------------------------------------
//-------------------------------------------------------------------------------

void TetSet::read(int _numTets, FILE *fptr)  
{

  char line[MAXLINE];

  numTets = _numTets;

  tets = new Tet[numTets];

  for (int i=0; i<numTets; ++i) {

    fgets(line, MAXLINE, fptr);
    sscanf(line, "%d %d %d %d", &tets[i][0], &tets[i][1], &tets[i][2], &tets[i][3]);
    
    for (int j=0; j<4; ++j) tets[i][j] -= 1;
    
  }

  fprintf(stdout, "       %d tetrahedra\n", numTets);

}

//-------------------------------------------------------------------------------

void TetSet::readXpost(int _numTets, FILE *fptr)  
{

  char line[MAXLINE];

  numTets = _numTets;

  tets = new Tet[numTets];

  for (int i=0; i<numTets; ++i) {

    fgets(line, MAXLINE, fptr);
    sscanf(line, "%*s %*s %d %d %d %d", &tets[i][0], &tets[i][1], &tets[i][2], &tets[i][3]);
    
    for (int j=0; j<4; ++j) tets[i][j] -= 1;
    
  }

  fprintf(stdout, "       %d tetrahedra\n", numTets);

}

//-------------------------------------------------------------------------------

void TetSet::read(int _numTets, int (*data)[4])  
{

  numTets = _numTets;

  tets = new Tet[numTets];

  for (int i=0; i<numTets; ++i)
    for (int j=0; j<4; ++j) 
      tets[i][j] = data[i][j];

}

//-------------------------------------------------------------------------------

void TetSet::numberEdges(EdgeSet &edges)	
{

  for (int i=0; i<numTets; ++i) 
    tets[i].numberEdges(edges);

}

//-------------------------------------------------------------------------------
/*
int Tet::pointInPolyhedron(NodeSet &nodes, double *Tmin, double *Tmax, double *pa)
{

  double eps = 1.;
  double pb[3] = {Tmax[0]+eps, pa[1], pa[2]};

  int n0 = 0;
  int n1 = 1;
  int n2 = 2;
  int n3 = 3;
  int na = 4;
  int nb = 5;

  Vec3D p0 = nodes[ nodeNum[0] ];
  Vec3D p1 = nodes[ nodeNum[1] ];
  Vec3D p2 = nodes[ nodeNum[2] ];
  Vec3D p3 = nodes[ nodeNum[3] ];

  int cross = 0;
  cross += pierceTriangularFace(n0, p0, n1, p1, n2, p2, na, pa, nb, pb);
  cross += pierceTriangularFace(n0, p0, n2, p2, n3, p3, na, pa, nb, pb);
  cross += pierceTriangularFace(n0, p0, n3, p3, n1, p1, na, pa, nb, pb);
  cross += pierceTriangularFace(n1, p1, n3, p3, n2, p2, na, pa, nb, pb);

  if (cross % 2 == 0) return 0;
  else return 1;

}

//-------------------------------------------------------------------------------
// test if segment [pa, pb] pierces the face defined by the points p0, p1, p2

int pierceTriangularFace(int n0, double *p0, int n1, double *p1, int n2, double *p2, 
			 int na, double *pa, int nb, double *pb)
{

  printf("\n");
  printf("p0: %f %f %f\n",p0[0], p0[1], p0[2]);
  printf("p1: %f %f %f\n",p1[0], p1[1], p1[2]);
  printf("p2: %f %f %f\n",p2[0], p2[1], p2[2]);
  printf("pa: %f %f %f\n",pa[0], pa[1], pa[2]);
  printf("pa: %f %f %f\n",pb[0], pb[1], pb[2]);

  double vola = orient3drobust(n0, p0, n1, p1, n2, p2, na, pa);
  double volb = orient3drobust(n0, p0, n1, p1, n2, p2, nb, pb);

  if (vola*volb > 0.0) return 0;
    
  double vol1 = orient3drobust(na, pa, n1, p1, n2, p2, nb, pb);
  double vol2 = orient3drobust(na, pa, n0, p0, n1, p1, nb, pb);
  double vol3 = orient3drobust(na, pa, n2, p2, n0, p0, nb, pb);

  if ((vol1 < 0.0) && (vol2 < 0.0) && (vol3 < 0.0)) return 1;
  if ((vol1 > 0.0) && (vol2 > 0.0) && (vol3 > 0.0)) return 1;

  return 0;

}

//-------------------------------------------------------------------------------

double orient3drobust(int na, double *pa, int nb, double *pb, 
		      int nc, double *pc, int nd, double *pd)
{

  double xa[3], xb[3], xc[3], xd[3];
  double epsa[3], epsb[3], epsc[3], epsd[3], eps = 0.95;

  double vol6 = orient3d(pa, pb, pc, pd);

  if (vol6 == 0.0) {

    printf(" *** WARNING: tie-breaking\n");

    int j;
    for (j=0; j<3; ++j) {
      epsa[j] = pow(eps, pow(2.0, double(na*3 - j - 1)));
      epsb[j] = pow(eps, pow(2.0, double(nb*3 - j - 1)));
      epsc[j] = pow(eps, pow(2.0, double(nc*3 - j - 1)));
      epsd[j] = pow(eps, pow(2.0, double(nd*3 - j - 1)));
    }

    for (j=0; j<3; ++j) {
      xa[j] = pa[j] + epsa[j];
      xb[j] = pb[j] + epsb[j];
      xc[j] = pc[j] + epsc[j];
      xd[j] = pd[j] + epsd[j];
    }

    vol6 = orient3d(xa, xb, xc, xd);

    //    if (vol6 == 0.0) {printf(" *** ERROR: tie-breaking failed\n");

    printf("%.30g %.30g %.30g\n",pa[0],pa[1],pa[2]);
    printf("%.30g %.30g %.30g\n",xa[0],xa[1],xa[2]);
    printf(">> %.30g %.30g %.30g\n",epsa[0],epsa[1],epsa[2]);
    printf("%.30g %.30g %.30g\n",pb[0],pb[1],pb[2]);
    printf("%.30g %.30g %.30g\n",xb[0],xb[1],xb[2]);
    printf(">> %.30g %.30g %.30g\n",epsb[0],epsb[1],epsb[2]);
    printf("%.30g %.30g %.30g\n",pc[0],pc[1],pc[2]);
    printf("%.30g %.30g %.30g\n",xc[0],xc[1],xc[2]);
    printf(">> %.30g %.30g %.30g\n",epsc[0],epsc[1],epsc[2]);
    printf("%.30g %.30g %.30g\n",pd[0],pd[1],pd[2]);
    printf("%.30g %.30g %.30g\n",xd[0],xd[1],xd[2]);
    printf(">> %.30g %.30g %.30g\n",epsd[0],epsd[1],epsd[2]);
    exit(1);

  }

  return vol6;

}

//-------------------------------------------------------------------------------

void Tet::computeCentroid(const NodeSet &nodes, Vec3D &centroid)
{

  centroid = 0.25 * ( nodes[ nodeNum[0] ] + nodes[ nodeNum[1] ] +
		      nodes[ nodeNum[2] ] + nodes[ nodeNum[3] ] );
}
*/
//-------------------------------------------------------------------------------

